package milab.lab03.ControllerAgent;

import jade.core.Agent;
import jade.core.AID;
import jade.core.behaviours.*;
import jade.lang.acl.ACLMessage;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;

public class ControllerAgent extends Agent {

  // The location of the knowledge base (a CLP file) of the agent
  private String knowledgeBase;
  // The list of known seller agents
  private AID[] sensorAgents;

  // Put agent initializations here
	protected void setup() {
	  // Printout a welcome message
	  System.out.println("Hello! Controller-agent " + getAID().getName() + " is ready");

	  // Get the location of the knowledge base and the type of sensors as a start-up argument
	  Object[] args = getArguments();

	  if (args != null && args.length > 0) {

	    knowledgeBase = (String)args[0];
	    System.out.println("Knowledge base is " + knowledgeBase);

	    // Add a TickerBehaviour that schedules a query to sensor agents after a given time (ms)
	    // This time corresponds to a kind of sampling rate (of the Environment)
	    // The higher the sampling rate and so the lower the interval between consecutive samplings, the more robust the agent will be
	    // to the dynamics of the Environment (e.g. interim changes - independent from the agent - to the value of measured parameters).
	    // But really this robustness depends on the time the agent spends on computation (after perception) and thus on how fast it reacts.
	    // Sampling time shouldn't be lower than the time of "thinking". If it is, then it may lead to inconsistency. Facts may be inserted or
	    // retracted to the KB which affect the ongoing reasoning. So let's assume, that the sampling interval is at least the computation time!
	    // If the computation between perception and action takes too long, we'll have the problem of knowledge inconsistency (the agent 
	    // will have false beliefs), and so may execute an inappropriate action. This action then may lead to unexpected effects, or even no
	    // effect at all (because of the structure of the environment)! Suppose that a parameter - according to the structure of the Env. -
	    // may only change from 1 to 2, or from 2 to 3. Lets say the agent observes the value of this parameter is 1, and wants to change it
	    // to 2. But while the agent was "thinking", and just before the next measurement of the parameter, an other agent changes the value
	    // from 1 to 2!! Our agent ignores this fact (because it couldn't know about it), and thus tries to change the value of the parameter
	    // from 1 to 2. But the value is 2, so the agent can't change it: the action has no effect... -------- Moreover, actions in (quasi)
	    // non-deterministic environments may have unexpected effects independently from the fact if the agent's beliefs are true or not.
	    // Note also: measurement/observation of parameters (not only according to W. Heisenberg) is a kind of intervention (action) in the
	    // Environment. It may even affect/modify values of parameters (depending on the Environment, the interdependence of parameters, etc).
	    // And let's not forget, that the agent is also part of its' environment, and so may measure or have actions directly affecting itself!
	    addBehaviour(new TickerBehaviour(this, 5000) {

	      protected void onTick() {

			System.out.println("Trying to locate sensors");// of type " + sensorType);
			// Update the list of sensor agents
			DFAgentDescription template = new DFAgentDescription();
			ServiceDescription sd = new ServiceDescription();
			sd.setType("general-sensor");
			template.addServices(sd);

			try {

				DFAgentDescription[] result = DFService.search(myAgent, template);
          		System.out.println("Found the following sensor agents:");
          		sensorAgents = new AID[result.length];

          		for (int i = 0; i < result.length; ++i) {

            		sensorAgents[i] = result[i].getName();
	          		System.out.println(sensorAgents[i].getName());

            	}

			} catch (FIPAException fe) {

				fe.printStackTrace();

			}

			// Perform the request
	        myAgent.addBehaviour(new RequestPercepts());
	        //myAgent.addBehaviour(new BasicJessBehaviour(myAgent, knowledgeBase, 1));
	        
	      }

	    });
	    
	    // Add the intelligent rule-based behaviour to the agent (reacting to arriving messages, i.e. information)
	    addBehaviour(new ControllerJessBehaviour(this, knowledgeBase));

	  } else {

	    // Make the agent terminate
	    System.out.println("No knowledge base specified. Terminating...");
	    doDelete();

	  }

	}

	// Put agent clean-up operations here
	protected void takeDown() {

		// Printout a dismissal message
		System.out.println("Controller-agent " + getAID().getName() + " terminating.");

	}

	private class RequestPercepts extends OneShotBehaviour {
			
		public void action() {
	  			    
			// Send the percept-request to all sensor agents
			ACLMessage msg = new ACLMessage(ACLMessage.REQUEST);

		  	for (int i = 0; i < sensorAgents.length; ++i) {

		  		msg.addReceiver(sensorAgents[i]);

			}

		 	msg.setContent("");		      
		 	msg.setConversationId("sensor-info");
		  	msg.setReplyWith("req"+System.currentTimeMillis()); // Unique value
		   	send(msg);

		}

	}  // End of inner class RequestPercepts

}
